﻿using HeuristicLab.Operators;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Optimization;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Algorithms.SimulatedAnnealing
{
    [Item("Fixed Reheater", "Increases temperature by a fixed amount, each time n consecutive solutions are not accepted.")]
    [StorableClass]
    public class FixedReheater : SingleSuccessorOperator, ISimulatedAnnealingHeatingStrategy
    {
        #region Strings
        private const string AmountName = "Amount";
        private const string ThresholdName = "Threshold";
        private const string TemperatureName = "Temperature";
        private const string AcceptanceMemoryName = "AcceptanceMemory";


        #endregion
        #region Parameters
        private ValueParameter<DoubleValue> AmountParameter
        {
            get { return (ValueParameter<DoubleValue>)Parameters[AmountName]; }
        }
        private ValueParameter<IntValue> ThresholdParameter
        {
            get { return (ValueParameter<IntValue>)Parameters[ThresholdName]; }
        }
        public ILookupParameter<DoubleValue> TemperatureParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[TemperatureName]; }
        }
        public ILookupParameter<ItemList<BoolValue>> AcceptanceMemoryParameter
        {
            get { return (ILookupParameter<ItemList<BoolValue>>)Parameters[AcceptanceMemoryName]; }
        }

        #endregion
        #region Properties
        public DoubleValue Amount
        {
            get { return AmountParameter.Value; }
            set { AmountParameter.Value = value; }
        }
        public IntValue AcceptanceThreshold
        {
            get { return ThresholdParameter.Value; }
            set { ThresholdParameter.Value = value; }
        }
        #endregion

        public FixedReheater() : base()
        {
            #region Create parameters
            Parameters.Add(new ValueParameter<DoubleValue>(AmountName, "The amount the temperature gets increased with every reheat.", new DoubleValue(20d)));
            Parameters.Add(new ValueParameter<IntValue>(ThresholdName, "Amount of consecutive rejected solutions that will trigger a reheat.", new IntValue(5)));
            Parameters.Add(new LookupParameter<DoubleValue>(TemperatureName, "The current temperature."));
            Parameters.Add(new LookupParameter<ItemList<BoolValue>>(AcceptanceMemoryName, "Memorizes the last N acceptance decisions."));
            #endregion
        }

        [StorableConstructor]
        protected FixedReheater(bool deserializing) : base(deserializing) { }
        protected FixedReheater(FixedReheater original, Cloner cloner) : base(original, cloner) { }


        public override IDeepCloneable Clone(Cloner cloner)
        {
            return new FixedReheater(this, cloner);
        }

        public void Parameterize() { }

        public bool ShouldReheat(bool cooling, ItemList<BoolValue> acceptances) {
            if (acceptances.Count > ThresholdParameter.Value.Value) acceptances.RemoveAt(0);
            return acceptances.Count == ThresholdParameter.Value.Value && !acceptances.Any(x => x.Value);
        }
        public override IOperation Apply()
        {
            AcceptanceMemoryParameter.ActualValue.Clear();
            TemperatureParameter.ActualValue.Value += AmountParameter.Value.Value;
            return base.Apply();
        }
    }
}
